/* $Cambridge: hermes/src/prayer/prayer/cmd_spam.c,v 1.20 2005/08/10 09:48:34 dpc22 Exp $ */
/************************************************
 *    Prayer - a Webmail Interface              *
 ************************************************/

/* Copyright (c) University of Cambridge 2000 - 2002 */
/* See the file NOTICE for conditions of use and distribution. */

#include "prayer_session.h"

static void generate_form(struct session *session)
{
    struct request *request = session->request;
    struct account *account = session->account;
    struct buffer *b = request->write_buffer;
    char *s;

    /* Title bar */
    html_start(session, b, "Junk Email (Spam) Filtering");
    html_banner_toolbar(session, b, "spam", 0);

    /* Title line */
    bputs(b, "<h2 align=\"center\">Junk Email (Spam) Filtering</h2>" CRLF);

    bputs(b, ("<p>This option automatically files mail based on how much it "
              "looks like junk email (spam)."CRLF));

    bputs(b, ("<p>Each message is given a score which is higher if the "
              "message has more spam-like features."CRLF));

    bputs(b, ("You pick a threshold score above which messages are "
              "filtered to a folder named &quot;spam&quot;,"));
    bputs(b, ("which you will have to check periodically "
	      "(once a week, say) for misfiled email "
	      "and to delete all the spam."CRLF));

    bputs(b, ("The lower the threshold the more likely it is to "
              "move real email to your spam folder."CRLF));

    bputs(b, ("Try 10 at first, "
              "and adjust based on how well it works for you."CRLF));

    /* Form body */
    html_form_start(session, b, "POST", "spam");

    bputs(b, "<table>" CRLF);
    bputs(b, "<tr>" CRLF);

    if (account->spam_threshold_valid)
        s = string_itoa_tmp(account->spam_threshold);
    else
        s = "10";
    html_form_table_field(b, "Filter threshold:", "spam_threshold", s, 5);

    if (account->spam_enabled)
        bputs(b, "<td>(currently enabled)</td>"CRLF);
    else
        bputs(b, "<td>(currently disabled)</td>"CRLF);

    bputs(b, "</tr>" CRLF);

    if (session->sieved_server) {
        bputs(b, "<tr><td>");
        bprintf(b,
                "<input type=\"checkbox\" name=\"spam_purge\" value=\"1\"%s>",
                (account->spam_purge_enabled != 0) ? "checked" : "");
        bputs(b, " Purge after:");
        bputs(b, "</td><td>");
        bprintf(b, ("<input name=\"spam_purge_timeout\" "
                    "value=\"%d\" size=\"5\">"),
                account->spam_purge_timeout);
        bputs(b, "</td>");
        bputs(b, "<td>days</td></tr>"CRLF);

    }
    bputs(b, "</table>"CRLF);

    if (session->sieved_server) {
        bputs(b, "<table><tr><td>");
        bputs(b, "Whitelist (one address per line):");
        bputs(b, "</td><td>"CRLF);
        bprintf(b,
                "<textarea name=\"spam_whitelist\" rows=\"8\" cols=\"40\">"
                CRLF);
        if (account->spam_whitelist && account->spam_whitelist[0])
            html_quote_string(b, account->spam_whitelist);
        bputs(b, "</textarea>" CRLF);
        bputs(b, "</td></tr></table>"CRLF);
    }

    bputs(b, "<table>" CRLF);
    bputs(b, "<tr>" CRLF);
    html_form_table_submit(b, "sub_cancel", "Back to Options Screen");
    html_form_table_submit(b, "sub_disable", "Disable Filtering");
    html_form_table_submit(b, "sub_enable",
                           "Enable Filtering with this threshold");
    bputs(b, "</tr>" CRLF);
    bputs(b, "</table>" CRLF);
    html_form_end(b);
    html_end(b);
    response_html(request, 200);        /* Success */
}

static void generate_error(struct session *session)
{
    struct request *request = session->request;
    struct account *account = session->account;
    char *msg = account_fetch_message(account);
    struct buffer *b = request->write_buffer;

    /* Title bar */
    html_start(session, b, "Junk Email (Spam) filtering");
    html_banner_toolbar(session, b, "spam", 0);

    /* Title line */
    bputs(b, "<h2 align=\"center\">Junk Email (Spam) Filtering</h2>" CRLF);

    bputs(b, "<b>Error:</b> ");
    if (msg) {
        html_quote_string(b, msg);
        session_log(session, "[cmd_spam] %s", msg);
    } else {
        bputs(b, "Unable to check mail processing status with server");
    }
    bputs(b, "" CRLF);

    /* Form body */
    html_form_start(session, b, "POST", "spam");

    /* Cancel button */
    bputs(b, "<table>" CRLF);
    bputs(b, "<tr>" CRLF);
    html_form_table_submit(b, "sub_cancel", "Back to Options Screen");
    bputs(b, "</tr>" CRLF);
    bputs(b, "</table>" CRLF);
    html_form_end(b);

    html_end(b);
    response_html(request, 200);
}

/* Remove extraneous whitespace from each line in whitelist */

static char *
normalise_whitelist(char *list, struct pool *pool)
{
    struct buffer *b = buffer_create(pool, 64);
    char *tmp = pool_strdup(pool, list);         /* Scratch copy */
    char *s;

    while ((s=string_get_line(&tmp))) {
        if ((s=string_trim_whitespace(s)) && s[0]) {
            bputs(b, s);
            bputs(b, ""CRLF);
        }
    }
    return(buffer_fetch(b, 0, buffer_size(b), NIL));
}

/* Check that addresslist consists of one valid address per line
   (wildcard characters are allowed in addresses) */

static BOOL
validate_whitelist(struct session *session, char *list, struct pool *pool)
{
    char *tmp = pool_strdup(pool, list); /* Scratch copy */
    char *s;
    ADDRESS *addr;

    while ((s=string_get_line(&tmp))) {
        if ((s=string_trim_whitespace(s)) && s[0]) {
            addr = addr_parse(pool, s, "");

            if (!addr) {
                session_message(session, "Invalid address: %s", s);
                return(NIL);
            }
            if (addr->next) {
                session_message(session,
                                "Single address on each line please: %s", s);
                mail_free_address(&addr);
                return(NIL);
            }
            mail_free_address(&addr);
        }
    }
    return(T);
}

void cmd_spam(struct session *session)
{
    struct config  *config = session->config;
    struct request *request = session->request;
    struct assoc *h = NIL;
    struct account *account = session->account;
    char *s, *t;

    if (request->method != POST) {
        if (!account_mail_check(account, request->pool))
            generate_error(session);
        else
            generate_form(session);
        return;
    }

    /* Form post */
    request_decode_post(request);
    h = request->form;

    if (assoc_lookup(h, "sub_cancel")) {
        session_redirect(session, request, "manage");
        return;
    }

    s = assoc_lookup(h, "spam_threshold");

    if (!(s && string_isnumber(s) && (atoi(s) > 0))) {
        session_message(session,
                        "Threshold must be a number greater than zero");
        generate_form(session);
        return;
    }

    account->spam_enabled = assoc_lookup(h, "sub_enable") ? T : NIL;
    account->spam_threshold = atoi(s);
    account->spam_threshold_valid = T;

    /* Spam whitelist and purge currently only available to sieve users */
    if (session->sieved_server) {
        if ((t = assoc_lookup(h, "spam_whitelist"))) {
            string_strdup(&account->spam_whitelist,
                          normalise_whitelist(t, request->pool));

            if (!validate_whitelist(session, account->spam_whitelist,
                                    request->pool)) {
                generate_form(session);
                return;
            }
        } else
            string_strdup(&account->spam_whitelist, "");

        t = assoc_lookup(h, "spam_purge_timeout");
        if (!(t && string_isnumber(t) && (atoi(t) >= 0))) {
            session_message(session,
                            "Purge time must be a non-negative integer");
            generate_form(session);
            return;
        }

        account->spam_purge_timeout = atoi(t);
        account->spam_purge_enabled = assoc_lookup(h, "spam_purge") ? T : NIL;

        if (account->spam_purge_enabled) {
            session_log(session, "[cmd_spam] spam_purge enabled: %d days", 
                        account->spam_purge_timeout);
        } else {
            session_log(session, "[cmd_spam] spam_purge disabled");
        }
    }

    if (!account_mail_update(account, request->pool)) {
        char *msg = account_fetch_message(account);

        session_message(session, "Failed to update spam filter: %s", msg);
        session_log(session,
                    "[cmd_spam] Failed to update spam filter: %s", msg);
        return;
    }

    if (account->spam_enabled) {
        /* Create spam folder if it doesn't already exist */
        /* XXX Rather nasty: really need merged dircache/dirlist system */
        char *maildir = (config->maildir) ? (config->maildir) : "";
        struct dirlist *dirlist = dircache_fetch(session, maildir);
        char *spam_name = (maildir[0]) 
            ? pool_strcat(request->pool, maildir, "/spam")
            : "spam";

        while (dirlist && (strcmp(dirlist->name, "spam") != 0))
            dirlist = dirlist->next;

        if (!dirlist) {
            if (ml_create(session, session->stream,
                          session_mailbox(session, request->pool, spam_name)))
                dircache_add(session->dircache, spam_name, NIL);
            else
                session_message(session, "Failed to create spam folder");
        }
    }

    session_message(session, "Updated junk email filtering");
    session_redirect(session, request, "manage");
}
